from Acquisition import aq_inner
from collective.solr.browser.facets import FacetMixin
from collective.solr.browser.facets import facetParameters
from collective.solr.browser.facets import param
from ftw.solr.indexers import site_area as site_area_indexer
from plone import api
from plone.app.layout.viewlets import common
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot


class SearchBoxViewlet(common.SearchBoxViewlet, FacetMixin):

    def initial_site_area_filter(self):
        if 'facet.field' in self.request.form:
            return None
        obj = aq_inner(self.context)

        if IPloneSiteRoot.providedBy(api.portal.get_navigation_root(obj)):
            return None
        else:
            return site_area_indexer(obj)()

    def hiddenfields(self):
        facets, dependencies = facetParameters(self)
        facets = list(facets)
        queries = param(self, 'fq')
        site_area = self.initial_site_area_filter()
        if site_area:
            queries = ['site_area:"%s"' % site_area]
            if 'site_area' in facets:
                facets.remove('site_area')

        return self.hidden(facets=facets, queries=queries)



from Products.CMFCore.utils import getToolByName
from ftw.solr.interfaces import IFtwSolrLayer
from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from plone.browserlayer.utils import registered_layers
from unittest2 import TestCase


class TestInstallation(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def test_ftw_solr_profile_installed(self):
        portal = self.layer['portal']
        portal_setup = getToolByName(portal, 'portal_setup')

        version = portal_setup.getLastVersionForProfile(
            'ftw.solr:default')
        self.assertNotEqual(version, None)
        self.assertNotEqual(version, 'unknown')

    def test_collective_solr_profile_installed(self):
        portal = self.layer['portal']
        portal_setup = getToolByName(portal, 'portal_setup')

        version = portal_setup.getLastVersionForProfile(
            'collective.solr:default')
        self.assertNotEqual(version, None)
        self.assertNotEqual(version, 'unknown')

    def test_request_layer_active(self):
        layers = registered_layers()
        self.assertIn(IFtwSolrLayer, layers)



from os.path import dirname, join
from ftw.solr import tests


def getData(filename):
    """ return a file object from the test data folder """
    filename = join(dirname(tests.__file__), 'data', filename)
    return open(filename, 'r').read()



from ftw.upgrade import UpgradeStep


class RegisterJavascriptForAccessiblityHandlingOfTheFacets(UpgradeStep):
    """Register javascript for accessiblity handling of the facets.
    """

    def __call__(self):
        self.install_upgrade_profile()



from unittest import TestCase
from ftw.solr.flare import PloneFlare


class TestPloneFlare(TestCase):

    def test_unicode_string_attribute_returned_as_byte_string(self):
        flare = PloneFlare(dict(ustring=u'Foo'))
        self.assertTrue(isinstance(flare.ustring, str))

    def test_byte_string_attribute_returned_as_byte_string(self):
        flare = PloneFlare(dict(string='Foo'))
        self.assertTrue(isinstance(flare.string, str))



from ftw.solr.patches.utils import isSimpleSearch
from ftw.solr.patches.utils import isSimpleTerm
from unittest import TestCase


class TestIsSimpleSearch(TestCase):

    def test_is_simple_search_considers_anyting_with_quotes_not_simple(self):
        self.assertTrue(isSimpleSearch('foo'))
        self.assertTrue(isSimpleSearch('foo bar'))
        self.assertFalse(isSimpleSearch('"bar"'))
        self.assertFalse(isSimpleSearch('foo "bar"'))
        self.assertFalse(isSimpleSearch('foo"bar'))

    def test_simple_searches_may_contain_dots(self):
        self.assertTrue(isSimpleSearch('foo.bar'))
        self.assertTrue(isSimpleSearch('foo.bar baz'))


class TestIsSimpleTerm(TestCase):

    def test_simple_terms_may_contain_dots(self):
        self.assertTrue(isSimpleTerm('foo.bar'))

    def test_simple_terms_may_contain_digits(self):
        self.assertTrue(isSimpleTerm('foo7bar'))
        self.assertTrue(isSimpleTerm('foo7'))



# -*- coding: utf-8 -*-
from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from unittest2 import TestCase
from zope.component import getMultiAdapter
from zope.event import notify
from zope.traversing.interfaces import BeforeTraverseEvent
from collective.solr.flare import PloneFlare
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from zope.component import getUtility
from plone.registry.interfaces import IRegistry
from ftw.solr.interfaces import ISearchSettings
from collective.solr.parser import SolrResponse
from plone.app.contentlisting.interfaces import IContentListingObject


class TestSearchView(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def test_breadcrumbs(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        setRoles(portal, TEST_USER_ID, ['Manager'])
        f1 = portal[portal.invokeFactory('Folder', 'f1', title=u"Folder 1")]
        f2 = f1[f1.invokeFactory('Folder', 'f2', title=u"Folder 2")]
        f3 = f2[f2.invokeFactory('Folder', 'f3', title=u"Folder 3")]

        flare = PloneFlare(portal)
        flare['path_string'] = '/plone/f1/f2/f3'
        flare = IContentListingObject(flare)
        view = getMultiAdapter((f3, request), name=u'search')
        breadcrumbs = view.breadcrumbs(flare)
        self.assertEquals(len(breadcrumbs), 2)
        self.assertEquals(breadcrumbs[0]['Title'], 'Folder 1')
        self.assertEquals(breadcrumbs[0]['absolute_url'], f1.absolute_url())
        self.assertEquals(breadcrumbs[1]['Title'], 'Folder 2')
        self.assertEquals(breadcrumbs[1]['absolute_url'], f2.absolute_url())

    def test_path_based_breadcrumbs(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        # Enable path based breadcrumbs
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISearchSettings)
        settings.path_based_breadcrumbs = True

        flare = PloneFlare(portal)
        flare['path_string'] = '/plone/path/to/object'
        flare = IContentListingObject(flare)
        view = getMultiAdapter((portal, request), name=u'search')
        breadcrumbs = view.breadcrumbs(flare)
        self.assertEquals(len(breadcrumbs), 2)
        self.assertEquals(breadcrumbs[0]['Title'], 'path')
        self.assertEquals(breadcrumbs[0]['absolute_url'],
            portal.absolute_url() + '/path')
        self.assertEquals(breadcrumbs[1]['Title'], 'to')
        self.assertEquals(breadcrumbs[1]['absolute_url'],
            portal.absolute_url() + '/path/to')

    def test_breadcrumbs_cropping(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        # Enable path based breadcrumbs
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISearchSettings)
        settings.path_based_breadcrumbs = True

        flare = PloneFlare(portal)
        flare['path_string'] = '/plone/a/long/path/to/an/object'
        flare = IContentListingObject(flare)
        view = getMultiAdapter((portal, request), name=u'search')
        breadcrumbs = view.breadcrumbs(flare)
        self.assertEquals(len(breadcrumbs), 4)
        self.assertEquals(breadcrumbs[0]['Title'], 'a')
        self.assertEquals(breadcrumbs[1]['Title'], u'…')
        self.assertEquals(breadcrumbs[2]['Title'], 'to')
        self.assertEquals(breadcrumbs[3]['Title'], 'an')

        settings.max_breadcrumbs = 5
        breadcrumbs = view.breadcrumbs(flare)
        self.assertEquals(len(breadcrumbs), 5)
        self.assertEquals(breadcrumbs[0]['Title'], 'a')
        self.assertEquals(breadcrumbs[1]['Title'], 'long')
        self.assertEquals(breadcrumbs[2]['Title'], 'path')
        self.assertEquals(breadcrumbs[3]['Title'], 'to')
        self.assertEquals(breadcrumbs[4]['Title'], 'an')

    def test_calling_search_view(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        view = getMultiAdapter((portal, request), name=u'search')

        # Calling the @@search view without parameters shouldn't fail
        view()


    def test_filter_query_respecting_navroot(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISearchSettings)
        view = getMultiAdapter((portal, request), name=u'search')

        # When respect_navroot is disabled...
        settings.respect_navroot = False

        query = view.filter_query({'SearchableText': 'ham'})
        self.assertEquals(query.get('path'), None,
            "No path filter should be added if there wasn't one already")

        request.form.update({'path': '/foo'})
        query = view.filter_query({'SearchableText': 'ham'})
        self.assertEquals(query.get('path'), '/foo',
            'Existing path filter should remain unchanged')

        # When respect_navroot is enabled...
        settings.respect_navroot = True

        del request.form['path']
        query = view.filter_query({'SearchableText': 'ham'})
        self.assertEquals(query.get('path'), '/plone',
            'Search should be constrained to navroot')

        request.form.update({'path': '/foo'})
        query = view.filter_query({'SearchableText': 'ham'})
        self.assertEquals(query.get('path'), '/foo',
            'Existing path filter should remain unchanged')


    def test_suggestions(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        request.form.update({'SearchableText': 'bidlung', })
        view = getMultiAdapter((portal, request), name=u'search')
        view.solr_response = SolrResponse()
        view.solr_response.spellcheck = {}
        view.solr_response.spellcheck['suggestions'] = {
            'bidlung': {'endOffset': 246,
                        'numFound': 5,
                        'origFreq': 1,
                        'startOffset': 239,
                        'suggestion': [{'freq': 2704, 'word': 'bildung'},
                                       {'freq': 1, 'word': 'bidlungs'},
                                       {'freq': 1, 'word': 'bidung'},
                                       {'freq': 561, 'word': 'bildungs'},
                                       {'freq': 233, 'word': 'bislang'}]},
            'platform': {'endOffset': 336,
                         'numFound': 5,
                         'origFreq': 9,
                         'startOffset': 328,
                         'suggestion': [{'freq': 557, 'word': 'plattform'},
                                        {'freq': 2, 'word': 'platforma'},
                                        {'freq': 2, 'word': 'platforme'},
                                        {'freq': 2, 'word': 'platforms'},
                                        {'freq': 7, 'word': 'plateforme'}]},
            'correctlySpelled': False,
        }

        suggestions = view.suggestions()
        self.assertEquals(suggestions[0][0], 'bildung')
        self.assertEquals(suggestions[0][1], '&SearchableText=bildung')


    def test_suggestions_querystring_with_list_parameter(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        request.form.update({'SearchableText': 'bidlung',
                             'facet.field': ['portal_type', 'review_state']})
        view = getMultiAdapter((portal, request), name=u'search')
        view.solr_response = SolrResponse()
        view.solr_response.spellcheck = {}
        view.solr_response.spellcheck['suggestions'] = {
            'bidlung': {'endOffset': 246,
                        'numFound': 5,
                        'origFreq': 1,
                        'startOffset': 239,
                        'suggestion': [{'freq': 2704, 'word': 'bildung'},
                                       {'freq': 1, 'word': 'bidlungs'},
                                       {'freq': 1, 'word': 'bidung'},
                                       {'freq': 561, 'word': 'bildungs'},
                                       {'freq': 233, 'word': 'bislang'}]},
            'correctlySpelled': False,
        }

        suggestions = view.suggestions()
        self.assertEquals('bildung', suggestions[0][0])
        self.assertIn('&facet.field=portal_type&facet.field=review_state',
                      suggestions[0][1])
        self.assertIn('&SearchableText=bildung', suggestions[0][1])

    def test_suggestions_without_solr_response(self):
        portal = self.layer['portal']
        request = self.layer['request']

        # Setup browser layers
        notify(BeforeTraverseEvent(portal, request))

        view = getMultiAdapter((portal, request), name=u'search')
        self.assertEquals([], view.suggestions())



from plone.app.contentlisting import catalog
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.layout.icons.interfaces import IContentIcon
from zope.interface import implements
from zope.component import queryMultiAdapter
from zope.component.hooks import getSite
from DateTime import DateTime
import urllib

timezone = DateTime().timezone()


class SolrContentListingObject(catalog.CatalogContentListingObject):

    implements(IContentListingObject)

    def __init__(self, flare):
        self._brain = flare
        self._cached_realobject = None
        self.request = flare.request
        self.context = getSite()

    def __repr__(self):
        return "<ftw.solr.contentlisting." + \
               "SolrContentListingObject instance at %s>" % (self.getPath(), )

    def EffectiveDate(self, zone=timezone):
        return self._brain.effective.toZone(zone).ISO8601()

    def ExpirationDate(self, zone=timezone):
        return self._brain.expires.toZone(zone).ISO8601()

    def getIcon(self):
        return queryMultiAdapter((self.context, self.request,
                                  self._brain), interface=IContentIcon)()

    def appendViewAction(self):
        if self.is_external():
            return ''
        return super(SolrContentListingObject, self).appendViewAction()

    def is_external(self):
        if (getattr(self._brain, 'getRemoteUrl', None) and
                self._brain.getRemoteUrl.startswith('http')):
            return True
        return False

    def result_url(self):
        search_term = self.request.form.get('SearchableText', '')
        url = self._brain.getURL()
        anchor = None
        if '#' in url:
            url, anchor = url.split('#')
        url = url + self.appendViewAction()
        if search_term:
            url = url + '?searchterm=' + urllib.quote(search_term)
        if anchor:
            url = url + '#' + anchor
        return url



from collective.indexing.queue import getQueue
from plone.indexer.interfaces import IIndexer
from Products.Archetypes.config import TOOL_NAME
from Products.Archetypes.utils import isFactoryContained
from Products.CMFCore.interfaces import ICatalogTool
from Products.CMFCore.interfaces._content import ICatalogAware
from Products.CMFCore.utils import getToolByName
from zope.component import queryMultiAdapter
import logging


logger = logging.getLogger('ftw.solr')


def is_index_value_equal(old, new):
    """Compares an old and a new index value.
    If the value is of type list, it is compared as set since this usually is
    used in a KeywordIndex, where the order is not relevant.
    """
    if type(old) != type(new):
        return False
    if isinstance(old, list):
        return set(old) == set(new)
    return old == new


def is_index_up_to_date(catalog, obj, index_name):
    """Checks, whether the passed index (`index_name`) of the passed object (`obj`)
    is up to date on the passed catalog (`catalog`).
    """

    indexer = queryMultiAdapter((obj, catalog), IIndexer, name=index_name)
    if indexer is None:
        indexer = getattr(obj, index_name, None)

    if indexer is None:
        # We cannot re-generate the index data, therfore we
        # act as if it is outdated
        return False

    else:
        path = '/'.join(obj.getPhysicalPath())
        rid = catalog.getrid(path)
        if rid is None:
            # the object was not yet indexed
            return False

        indexed_values = catalog.getIndexDataForRID(catalog.getrid(path))
        value_before = indexed_values.get(index_name, None)
        value_after = indexer()
        if not is_index_value_equal(value_before, value_after):
            return False

    return True


def recursive_index_security(catalog, obj):
    """This function reindexes the security indexes for an object in a specific catalog
    recursively.
    It does this by walking down the tree and checking on every object whether a
    the security indexes are already up to date.
    If the security indexes are already up to date, it stops walking down the tree.

    The aim of stopping to walk down is to improve the performance drastically on
    large trees.
    The expectation is that if a children is up to date but the parent wasnt, the
    reason is usally that the children does not inherit the relevant values (for
    example it does not inherit the View permission) and thus the grand-children will
    also not change, so we can abort wakling down the path.
    """
    indexes_to_update = []

    # Since objectValues returns all objects, including placefulworkflow policy
    # objects, we have to check if the object is Catalog aware.
    if not ICatalogAware.providedBy(obj):
        return

    for index_name in obj._cmf_security_indexes:
        if not is_index_up_to_date(catalog, obj, index_name):
            indexes_to_update.append(index_name)

    if len(indexes_to_update) > 0:
        obj.reindexObject(idxs=indexes_to_update)

        # We assume that if the parent is up to date, all children are too.
        # This basically only walks into the tree untill an object is up to date -
        # usually because it does not inherit the security relevant thigns - and then
        # stops.
        for subobj in obj.objectValues():
            recursive_index_security(catalog, subobj)


def ftw_solr_CatalogMultiplex_reindexObjectSecurity(self, skip_self=False):
    """update security information in all registered catalogs.
    """
    if isFactoryContained(self):
        return
    at = getToolByName(self, TOOL_NAME, None)
    if at is None:
        return

    catalogs = [c for c in at.getCatalogsByType(self.meta_type)
                if ICatalogTool.providedBy(c)]

    for catalog in catalogs:
        recursive_index_security(catalog, self)


def ftw_solr_CatalogAware_reindexObjectSecurity(self, skip_self=False):
    """ Reindex security-related indexes on the object.
    """
    catalog = self._getCatalogTool()
    if catalog is None:
        return

    s = getattr(self, '_p_changed', 0)
    recursive_index_security(catalog, self)
    if s is None:
        self._p_deactivate()



from zope.interface import Interface
from zope import schema
from ftw.solr import _


class IFtwSolrLayer(Interface):
    """Browser layer for ftw.solr
    """

class ILiveSearchSettings(Interface):

    grouping = schema.Bool(
        title=_(u'Enable Grouping in LiveSearch'),
        description=_(u'If enabled, livesearch results are grouped by portal '
                       'type.'),
        default=False,
    )

    group_by = schema.List(
        title=_(u'Groups'),
        description=_(u'Specify a list of portal types by which livesearch '
                       'results are grouped. Groups are shown in the provided '
                       'order. Results that do not match a portal type in the '
                       'given list are shown at the bottom.'),
        default=[u'Folder', u'Document', u'News Item', u'Event', u'File',
                 u'Image'],
        value_type=schema.TextLine(),
        required=False,
    )

    group_search_limit = schema.Int(
        title=_(u'Group Search Limit'),
        description=_(u'The maximum number of search results to consider for'
                       'grouping.'),
        default=1000,
    )

    group_limit = schema.Int(
        title=_(u'Group Limit'),
        description=_(u'The maximum number of search results per group.'),
        default=7,
        required=False,
    )


    max_title = schema.Int(
        title=_(u'Maximum Title Length'),
        default=40,
    )

    max_description = schema.Int(
        title=_(u'Maximum Description Length'),
        default=100,
    )


class ISearchSettings(Interface):

    path_based_breadcrumbs = schema.Bool(
        title=_(u'Use Path for Breadcrumbs'),
        default=False,
    )

    max_breadcrumbs = schema.Int(
        title=_(u'Max. Breadcrumbs'),
        default=3,
    )

    respect_navroot = schema.Bool(
        title=_(u'Respect Navigation Root'),
        description=_(u'Constrain searches to navigation root'),
        default=False,
    )

class IZCMLSolrConnectionConfig(Interface):
    """Solr connection settings configured through ZCML.
    """



from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.manager import SolrConnectionConfig
from collective.solr.manager import SolrConnectionManager
from collective.solr.search import Search
from collective.solr.tests.utils import getData, fakehttp
from ftw.solr.patches.search import buildQueryAndParameters
from unittest import TestCase
from zope.component import provideUtility


class TestBuildQuery(TestCase):

    def setUp(self):
        provideUtility(SolrConnectionConfig(), ISolrConnectionConfig)
        self.mngr = SolrConnectionManager()
        self.mngr.setHost(active=True)
        conn = self.mngr.getConnection()
        fakehttp(conn, getData('plone_schema.xml'))  # fake schema response
        self.mngr.getSchema()                       # read and cache the schema
        self.search = Search()
        self.search.manager = self.mngr

        # Patch buildQuery method
        self.search.buildQuery = buildQueryAndParameters.__get__(
            self.search,
            self.search.__class__)

    def tearDown(self):
        self.mngr.closeConnection()
        self.mngr.setHost(active=False)

    def build_query(self, *args, **kwargs):
        query = self.search.buildQuery(*args, **kwargs)[0]
        return ' '.join(sorted(query.values()))

    def test_simple_path_query(self):
        self.assertEquals('+path_parents:\\/spam\\/and\\/eggs',
                          self.build_query(path_parents='/spam/and/eggs'))

    def test_path_list_query(self):
        self.assertEquals('+path_parents:("\\/spam" OR "\\/eggs")',
                          self.build_query(path_parents=['/spam', '/eggs']))

    def test_slash_in_searchabletext(self):
        self.assertEquals('+SearchableText:spam\\/eggs',
                          self.build_query(SearchableText='spam/eggs'))

    def test_queries_with_path_and_depth_should_be_escaped(self):
        self.assertEquals(
            '+path_depth:[4 TO 6] AND +path_parents:\\/spam\\/and\\/eggs',
            self.build_query(
                path_parents=set(['+path_depth:[4 TO 6] AND '
                                  '+path_parents:/spam/and/eggs'])))



from collective.solr.indexer import SolrIndexProcessor
from collective.solr.interfaces import ISolrIndexQueueProcessor
from zope.interface import implements


# TODO: Remove whole file in next major release.


class FtwSolrIndexProcessor(SolrIndexProcessor):
    implements(ISolrIndexQueueProcessor)
